from datetime import datetime, timezone
from typing import Any, Dict

import requests
from jwt.algorithms import get_default_algorithms
from jwt.api_jwt import decode, decode_complete


class AppleLogin:
    @classmethod
    def decode_id_token(cls, token: str, verify_exp: bool = True) -> Dict[str, Any]:
        unverified = decode_complete(token, options={"verify_signature": False})

        signing_keys = requests.get("https://appleid.apple.com/auth/keys").json()
        for key in signing_keys["keys"]:
            if key["kid"] == unverified["header"]["kid"]:
                signing_key = key
                break

        algo = get_default_algorithms()["RS256"]  # type: ignore

        signing_key = algo.from_jwk(signing_key)
        verified = decode(
            token, key=signing_key, algorithms=["RS256"], options={"verify_aud": False}
        )

        if verify_exp:
            now = datetime.now(timezone.utc).timestamp()
            if now > verified["exp"]:
                raise RuntimeError("Apple id token expired")

        return verified



import json
from typing import Any, Optional, TypeVar, Union

import boto3
from botocore.exceptions import ClientError
from mongoengine import DoesNotExist
from mongoengine.document import BaseDocument  # ignore: type
from pydantic import BaseModel, ValidationError

from fastel import jwt

from . import exceptions


class ClientConfig(BaseModel):
    client_id: str
    client_secret: str


AuthType = TypeVar("AuthType", bound="BaseAuth")
ClientType = TypeVar("ClientType", bound="ClientConfig")


class Credential:
    def __init__(
        self,
        client: Union[ClientType, BaseDocument, None],
        user: Any = None,
        server_call: bool = False,
    ) -> None:
        self.client = client
        self.user = user
        self.server_call = server_call


class BaseAuth:
    client_model_class: BaseDocument = None
    error_status_code: int = 410

    def __init__(self, force: bool = False):
        self.force = force

    def __or__(self, other: AuthType) -> "BaseAuth":
        return OR(self, other)

    def __call__(
        self,
        client_id: Optional[str] = None,
        client_secret: Optional[str] = None,
        token: Optional[str] = None,
    ) -> Optional[Credential]:
        return self.verify(
            client_id=client_id, token=token, client_secret=client_secret
        )

    def verify(
        self,
        client_id: Optional[str] = None,
        client_secret: Optional[str] = None,
        token: Optional[str] = None,
    ) -> Optional[Credential]:
        raise NotImplementedError("verify() should be implemented")  # pragma: no cover


class ClientIdAuth(BaseAuth):
    def __init__(self, force: bool = False):
        super().__init__(force)
        if self.client_model_class is None:
            raise ValueError(
                "client_model_class should not be None"
            )  # pragma: no cover

    def verify(
        self,
        client_id: Optional[str] = None,
        client_secret: Optional[str] = None,
        token: Optional[str] = None,
    ) -> Optional[Credential]:
        if not client_id and self.force:
            raise exceptions.APIException(
                status_code=self.error_status_code,
                error="access_denied",
                detail="client_id is required",
            )
        if not client_id:
            return None
        try:
            client = self.client_model_class.objects.get(client_id=client_id)
        except self.client_model_class.DoesNotExist as exc:
            raise exceptions.APIException(
                status_code=self.error_status_code,
                error="access_denied",
                detail=f"{client_id} not found",
            ) from exc
        return Credential(client=client)


class JWBaseAuth(BaseAuth):
    expected_aud: Optional[str] = None

    def verify(
        self,
        client_id: Optional[str] = None,
        client_secret: Optional[str] = None,
        token: Optional[str] = None,
    ) -> Optional[Credential]:
        if not token and self.force:
            raise exceptions.APIException(
                status_code=self.error_status_code,
                error="permission_denied",
                detail="token is required",
            )
        if not token:
            return None

        try:
            decoded = jwt.decode_static(token)
        except Exception as exc:
            raise exceptions.APIException(
                status_code=self.error_status_code,
                error="permission_denied",
                detail=str(exc),
            )

        if self.expected_aud and decoded["aud"] != self.expected_aud:
            raise exceptions.APIException(
                status_code=self.error_status_code,
                error="aud_denied",
                detail=f"required aud: {self.expected_aud}",
            )

        if self.client_model_class:
            try:
                client = self.client_model_class.objects.get(client_id=decoded["aud"])
            except self.client_model_class.DoesNotExist as exc:
                raise exceptions.APIException(
                    status_code=self.error_status_code,
                    error="permission_denied",
                    detail=f"{client_id} not found",
                ) from exc
        else:
            client = None
        return Credential(client=client, user=decoded)


class ClientSecretAuth(BaseAuth):
    root_client_class: BaseDocument

    def __init__(self, force: bool = False):
        super().__init__(force)
        if self.client_model_class is None:
            raise ValueError(
                "client_model_class should not be None"
            )  # pragma: no cover

    def verify(
        self,
        client_id: Optional[str] = None,
        client_secret: Optional[str] = None,
        token: Optional[str] = None,
    ) -> Optional[Credential]:

        if self.force and (client_secret is None or client_id is None):
            raise exceptions.APIException(
                status_code=self.error_status_code,
                error="permission_denied",
                detail="client_secret is required",
            )

        if client_secret is None or client_id is None:
            return None

        try:
            client = self.client_model_class.objects.get(client_id=client_id)
            root = self.root_client_class.objects.get(client_id=client_id)
        except DoesNotExist as exc:
            raise exceptions.APIException(
                status_code=self.error_status_code,
                error="permission_denied",
                detail=f"{client_id} not found",
            ) from exc

        if root.client_secret != client_secret:
            raise exceptions.APIException(
                status_code=self.error_status_code,
                error="permission_denied",
                detail="client_sercet not match",
            )

        return Credential(client=client, server_call=True)


class OR(BaseAuth):
    def __init__(self, left: BaseAuth, right: BaseAuth):
        super().__init__()
        self.left = left
        self.right = right

    def verify(self, *args: Any, **kwargs: Any) -> Optional[Credential]:
        result = self.left(*args, **kwargs)
        if not result:
            result = self.right(*args, **kwargs)
        return result


class StaticClientAuth(BaseAuth):
    s3_client = boto3.client("s3", "ap-northeast-1")
    client_bucket: str
    prefix_key: str = ""
    client_model = ClientConfig

    def verify(
        self,
        client_id: Optional[str] = None,
        client_secret: Optional[str] = None,
        token: Optional[str] = None,
    ) -> Optional[Credential]:
        if not client_id and self.force:
            raise exceptions.APIException(
                status_code=self.error_status_code,
                error="access_denied",
                detail="client_id is required",
            )
        if not client_id:
            return None
        try:
            obj = self.s3_client.get_object(
                Bucket=self.client_bucket, Key=f"{self.prefix_key}/{client_id}.json"
            )
            client_dict = json.load(obj["Body"])
            root = self.client_model.validate(client_dict)
        except ClientError:
            raise exceptions.APIException(
                status_code=self.error_status_code,
                error="access_denied",
                detail=f"{client_id} not found",
            )
        except ValidationError as exc:
            raise exceptions.APIException(
                status_code=411,
                error="config_error",
                detail=exc.errors(),
            )

        return Credential(client=root)


class StaticSecretAuth(BaseAuth):
    s3_client = boto3.client("s3", "ap-northeast-1")
    client_bucket: str
    prefix_key: str = ""
    client_model = ClientConfig

    def verify(
        self,
        client_id: Optional[str] = None,
        client_secret: Optional[str] = None,
        token: Optional[str] = None,
    ) -> Optional[Credential]:

        if self.force and (client_secret is None or client_id is None):
            raise exceptions.APIException(
                status_code=self.error_status_code,
                error="permission_denied",
                detail="client_secret is required",
            )

        if client_secret is None or client_id is None:
            return None
        try:
            obj = self.s3_client.get_object(
                Bucket=self.client_bucket, Key=f"{self.prefix_key}/{client_id}.json"
            )
            client_dict = json.load(obj["Body"])
            root = self.client_model.validate(client_dict)
        except ClientError:
            raise exceptions.APIException(
                status_code=self.error_status_code,
                error="access_denied",
                detail=f"{client_id} not found",
            )
        except ValidationError as exc:
            raise exceptions.APIException(
                status_code=411,
                error="config_error",
                detail=exc.errors(),
            )

        if root.client_secret != client_secret:
            raise exceptions.APIException(
                status_code=self.error_status_code,
                error="permission_denied",
                detail="client_sercet not match",
            )

        return Credential(client=root, server_call=True)


class StaticJWTAuth(BaseAuth):
    s3_client = boto3.client("s3", "ap-northeast-1")
    client_bucket: str
    prefix_key: str = ""
    client_model: Optional[ClientConfig] = None
    expected_aud: Optional[str] = None

    def __init__(self, force: bool = False) -> None:
        self.force = force

    def verify(
        self,
        client_id: Optional[str] = None,
        client_secret: Optional[str] = None,
        token: Optional[str] = None,
    ) -> Optional[Credential]:
        if not token and self.force:
            raise exceptions.APIException(
                status_code=self.error_status_code,
                error="permission_denied",
                detail="token is required",
            )
        if not token:
            return None

        try:
            decoded = jwt.decode_static(token)
        except Exception as exc:
            raise exceptions.APIException(
                status_code=self.error_status_code,
                error="permission_denied",
                detail=str(exc),
            )

        if self.expected_aud and decoded["aud"] != self.expected_aud:
            raise exceptions.APIException(
                status_code=self.error_status_code,
                error="aud_denied",
                detail=f"required aud: {self.expected_aud}",
            )

        if not self.client_model:
            return Credential(client=None, user=decoded, server_call=False)

        client_id = decoded["aud"]
        try:
            obj = self.s3_client.get_object(
                Bucket=self.client_bucket, Key=f"{self.prefix_key}/{client_id}.json"
            )
            client_dict = json.load(obj["Body"])
            root = self.client_model.validate(client_dict)
        except ClientError:
            raise exceptions.APIException(
                status_code=self.error_status_code,
                error="access_denied",
                detail=f"{client_id} not found",
            )
        except ValidationError as exc:
            raise exceptions.APIException(
                status_code=411,
                error="config_error",
                detail=exc.errors(),
            )

        return Credential(client=root, server_call=True, user=decoded)



from typing import Dict

from fastapi import APIRouter
from fastapi.routing import BaseRoute  # type: ignore

_api_docs = {}


def init_docs(docs: Dict[str, str]) -> None:
    global _api_docs
    _api_docs = docs


def router_with_doc(router: APIRouter) -> APIRouter:
    route: BaseRoute
    for route in router.routes:
        method = next(iter(route.methods))
        path = route.path
        key = f"{method} {path}"
        print(key)

        doc = _api_docs.get(key, None)
        if doc:
            route.description = doc
        else:
            route.description = "No doc available"
    return router



from typing import Any, Dict


class SdkConfig:
    auth_host: str = ""
    client_id: str = ""
    client_secret: str = ""
    extra_config: Dict[str, Any] = {}

    @classmethod
    def put_configs(cls, **kwargs: Any) -> None:
        for key, value in kwargs.items():
            if not hasattr(cls, key):
                print("SdkConfig: no such key", key, "/", value)
            else:
                setattr(cls, key, value)

    @classmethod
    def get_config(cls, key: str, default_value: Any = None) -> Any:
        return getattr(cls, key, default_value)

    @classmethod
    def put_extras(cls, **kwargs: Any) -> None:
        for key, value in kwargs.items():
            cls.extra_config[key] = value

    @classmethod
    def get_extra(cls, key: str, default_value: Any = None) -> Any:
        return cls.extra_config.get(key, default_value)



import base64
import json
from calendar import timegm
from datetime import datetime
from typing import Any, Dict, Optional

import boto3


def kms_encode(
    payload: Dict[str, Any],
    key: str,
    headers: Optional[Dict[str, Any]] = None,
) -> str:
    def base64url_encode(input: bytes) -> bytes:
        return base64.urlsafe_b64encode(input).replace(b"=", b"")

    # Payload
    payload = payload.copy()
    for time_claim in ["exp", "iat", "nbf"]:
        # Convert datetime to a intDate value in known time-format claims
        if isinstance(payload.get(time_claim), datetime):
            payload[time_claim] = timegm(payload[time_claim].utctimetuple())

    json_payload = json.dumps(payload, separators=(",", ":")).encode("utf-8")

    # Header
    header = {"typ": "JWT", "alg": "RS256"}

    if headers:
        header.update(headers)

    json_header = json.dumps(header, separators=(",", ":")).encode()

    segments = []
    segments.append(base64url_encode(json_header))
    segments.append(base64url_encode(json_payload))

    # Segments
    signing_input = b".".join(segments)
    try:
        client = boto3.client("kms", "ap-northeast-1")
        resp = client.sign(
            KeyId=key,
            Message=signing_input,
            SigningAlgorithm="RSASSA_PKCS1_V1_5_SHA_256",
        )
        signature = resp["Signature"]
    except Exception:
        raise RuntimeError("[ERROR] fail to sign with KMS")

    segments.append(base64url_encode(signature))
    encoded_string = b".".join(segments)
    return encoded_string.decode("utf-8")



import random
from typing import Any, Tuple, Type

from .decode_static import prod_cert, stg_cert
from .jwt_payload import JWTPayloadBuilder
from .kms_encode import kms_encode as encode
from .pem_coder import pem_encode, pem_random_key


class JWTEncoder:
    payload_builder_class: Type[JWTPayloadBuilder]
    stage: str = "stg"

    def __init__(self, client: Any) -> None:
        self.client = client

    def encode(self, user: Any, pem: bool = False) -> str:
        builder = self.payload_builder_class(client=self.client, data=user)
        payload = builder.get_payload()

        if pem:
            kid, private_key = pem_random_key(self.stage)
            return pem_encode(
                payload,
                private_key,
                headers={"kid": kid},
            )
        try:
            from plugin import public_key

            is_plugin = True
        except Exception:
            is_plugin = False
        if self.stage == "stg":
            plugin_resp: Tuple[bool, Any] = (
                is_plugin and public_key.stg_cert() or (False, None)
            )
            if plugin_resp[0]:
                keys = plugin_resp[1]
            else:
                keys = stg_cert

        else:
            plugin_resp = is_plugin and public_key.prod_cert() or (False, None)
            if plugin_resp[0]:
                keys = plugin_resp[1]
            else:
                keys = prod_cert
        key = random.choice(keys["keys"])["kid"]
        return encode(payload, key=key, headers={"kid": key})
